use mlua::{Compiler, FromLua, Function, IntoLuaMulti, Lua, Table, UserData, Value};
use sha2::{Digest, Sha256};
use std::{
    collections::HashMap,
    hash::Hash,
    sync::{
        atomic::{AtomicU16, Ordering},
        Arc,
    },
};
use tokio::{
    sync::{mpsc, oneshot, Mutex, RwLock},
    time::{timeout, Duration},
};

pub type ProgramId = [u8; 32];

type Sender<I, R> = mpsc::Sender<(I, oneshot::Sender<R>)>;
type Receiver<I, R> = mpsc::Receiver<(I, oneshot::Sender<R>)>;

#[derive(Debug)]
pub struct Program<I, R>
where
    for<'lua> I: IntoLuaMulti<'lua> + Sync + Send + 'lua,
    for<'lua> R: FromLua<'lua> + Clone + Sync + Send + 'lua,
{
    opcodes: Vec<u8>,
    instances: Arc<AtomicU16>,
    running: Arc<AtomicU16>,
    sender: Sender<I, R>,
    receiver: Arc<Mutex<Receiver<I, R>>>,
}

pub enum DbState {
    Balances,
    Accounts,
    Transactions,
    Other(String),
}

impl From<String> for DbState {
    fn from(s: String) -> Self {
        match s.as_str() {
            "balances" => Self::Balances,
            "accounts" => Self::Accounts,
            "transactions" => Self::Transactions,
            _ => Self::Other(s),
        }
    }
}

pub struct MetaState(Vec<DbState>);

impl UserData for MetaState {
    fn add_methods<'lua, M: mlua::UserDataMethods<'lua, Self>>(methods: &mut M) {
        methods.add_method_mut("__index", |_, _this, (_, key): (Table<'lua>, String)| {
            _this.0.push(key.into());
            Ok(1)
        });
    }
}

impl<I, R> Program<I, R>
where
    for<'lua> I: IntoLuaMulti<'lua> + Sync + Send + 'lua,
    for<'lua> R: FromLua<'lua> + Clone + Sync + Send + 'lua,
{
    pub fn new(opcodes: Vec<u8>) -> Program<I, R> {
        let (sender, receiver) = mpsc::channel(100);
        Self {
            instances: Arc::new(AtomicU16::new(0)),
            running: Arc::new(AtomicU16::new(0)),
            receiver: Arc::new(Mutex::new(receiver)),
            opcodes,
            sender,
        }
    }

    /// Returns a new Lua VM and a list of all the global variables to be
    /// persisted and read from the storage engine. Since lua is a dynamic
    /// language, other state variables may be read/updated dynamically, which
    /// is fine, this list is just for the initial state and any potential
    /// optimization.
    fn get_lua_vm(bytecode: &[u8]) -> (Vec<DbState>, Lua) {
        let lua = Lua::new();
        lua.sandbox(true).unwrap();
        let globals = lua.globals();
        globals.set("require", Value::Nil).unwrap();

        lua.load(bytecode).exec().unwrap();

        drop(globals);

        (vec![], lua)
    }

    fn spawn(
        bytecode: Vec<u8>,
        instances: Arc<AtomicU16>,
        running: Arc<AtomicU16>,
        receiver: Arc<Mutex<Receiver<I, R>>>,
    ) {
        if instances.load(Ordering::Relaxed) > 10 {
            return;
        }

        instances.fetch_add(1, Ordering::Relaxed);
        let max_timeout = Duration::from_secs(30);

        tokio::task::spawn_blocking(move || {
            let (_, lua) = Self::get_lua_vm(&bytecode);
            let f: Function = lua.globals().get("add").unwrap();

            loop {
                if let Ok(mut queue) =
                    futures::executor::block_on(timeout(max_timeout, receiver.lock()))
                {
                    if let Ok(Some(args)) =
                        futures::executor::block_on(timeout(max_timeout, queue.recv()))
                    {
                        // drop queue lock to release the mutex so any other
                        // free VM can use it to listen for incoming messages
                        drop(queue);
                        running.fetch_add(1, Ordering::Relaxed);
                        let ret = f.call::<I, Value>(args.0).unwrap();
                        let _ = args.1.send(R::from_lua(ret, &lua).unwrap());
                        continue;
                    }
                }
                break;
            }

            println!("Lua is exiting");
            instances.fetch_sub(1, Ordering::Relaxed);
        });
    }

    pub async fn exec(&self, input: I) -> R {
        let (return_notifier, return_listener) = oneshot::channel();
        Self::spawn(
            self.opcodes.clone(),
            self.instances.clone(),
            self.running.clone(),
            self.receiver.clone(),
        );
        self.sender
            .send((input, return_notifier))
            .await
            .expect("valid");
        return_listener.await.expect("valid")
    }
}

#[derive(Debug)]
pub struct VirtualMachines<K: Hash + Eq> {
    vms: RwLock<HashMap<K, Program<(i64, i64), i64>>>,
}

impl<K: Hash + Eq> VirtualMachines<K> {
    pub fn new() -> Arc<Self> {
        Arc::new(Self {
            vms: RwLock::new(HashMap::new()),
        })
    }

    pub async fn register_program(&self, name: K, program: &str) -> (ProgramId, bool) {
        self.register_opcodes(name, Compiler::new().compile(program))
            .await
    }

    pub async fn exec(&self, id: &K) -> Option<i64> {
        if let Some(vm) = self.vms.read().await.get(id) {
            Some(vm.exec((22, 33)).await)
        } else {
            None
        }
    }

    pub async fn register_opcodes(&self, name: K, opcodes: Vec<u8>) -> (ProgramId, bool) {
        let mut vms = self.vms.write().await;

        let mut hasher = Sha256::new();
        hasher.update(&opcodes);
        (
            hasher.finalize().into(),
            vms.insert(name, Program::new(opcodes)).is_some(),
        )
    }

    pub async fn shutdown(&self) {
        let mut vms = self.vms.write().await;
        vms.clear();
    }
}

#[tokio::main]
async fn main() {
    use std::{sync::Arc, time::Instant};
    async fn do_loop(vms: Arc<VirtualMachines<String>>) {
        // Create N threads to execute the Lua code in parallel
        let num_threads = 400;
        let (tx, mut rx) = mpsc::channel(num_threads);
        for _ in 0..num_threads {
            let vm = vms.clone();
            let tx_clone = tx.clone();

            tokio::spawn(async move {
                let start_time = Instant::now();
                let result = vm.exec(&"foo".to_owned()).await;

                // Send the result back to the main thread
                let _ = tx_clone.send(result).await;

                let elapsed_time = Instant::now() - start_time;

                // Print the elapsed time in seconds and milliseconds
                println!(
                    "Elapsed time: {} seconds {} milliseconds",
                    elapsed_time.as_secs(),
                    elapsed_time.as_millis(),
                );
            });
        }

        drop(tx);

        loop {
            let result = rx.recv().await;
            if result.is_none() {
                break;
            }
            println!("Result: {:?}", result,);
        }
    }

    let vms = VirtualMachines::new();

    // Compile Lua code
    let code = r#"
        calls = 0
        pid = 0
        function add(a, b)
            calls = calls + 1
            print("Call from old " .. pid .. " "  .. calls)
            return a + b
        end
        print("hello world " .. pid)
    "#;

    let _ = vms.register_program("foo".to_owned(), code).await;
    do_loop(vms.clone()).await;

    let code = r#"
        calls = 0
        pid = 1
        function add(a, b)
            calls = calls + 1
            print("Call from new " .. pid .. " "  .. calls)
            return a + b
        end
        print("hello world " .. pid)
    "#;
    let y = vms.register_program("foo".to_owned(), code).await;
    println!("{} {:?}", "foo", y);
    tokio::time::sleep(Duration::from_secs(3)).await;

    do_loop(vms.clone()).await;

    vms.shutdown().await;

    tokio::time::sleep(Duration::from_secs(1)).await;
}
